%{
    #include<iostream>
    #include<string>
    #include<map>
#ifdef TARGET_PLATFORM_NATIVE
    #define os cout
#endif
#ifdef TARGET_PLATFORM_WASM
    #include <sstream>
    #include <stdio.h> // for fmemopen
    #include <emscripten.h>
    stringstream os;
#endif
    using namespace std;
    #define YYERROR_VERBOSE


    bool verbose = false;
    bool expandnames = false;
    bool lazy_evaluation = true;
    bool latex_output = false;
    bool silent = false;


    class LambdaTerm
    {
        protected:
            static int replacement_var_id;
        public:
            virtual bool contains_free_var(string v) = 0;
            virtual bool reducible() = 0;
            virtual LambdaTerm* reduce() = 0;
            virtual LambdaTerm* substitute(string, LambdaTerm*) = 0;
            virtual string ToString() = 0;
            virtual LambdaTerm* Copy() = 0;
            virtual void SetBound(string vname) = 0;

            LambdaTerm() {}
            virtual ~LambdaTerm() {};
    };
    int LambdaTerm::replacement_var_id = 1;

    map<string, LambdaTerm*> symbol_table;

    class Variable : public LambdaTerm 
    {
        public:
            string name;
            bool bound;
            bool contains_free_var(string v) {return v==name && !bound;}

            Variable(string n, bool bnd=false)
            {
                name=n;
                bound=bnd;
            }
            
            LambdaTerm* Copy()
            {
                return new Variable(name,bound);
            }

            string ToString()
            {
                if(!expandnames || bound)
                    return name;
                if(symbol_table.find(name) != symbol_table.end())
                    return symbol_table[name]->ToString();
                return name;
            }

            void SetBound(string vname)
            {
                if(name == vname)
                    bound = true;
            }

            LambdaTerm* substitute(string v, LambdaTerm *Term)
            {
                if(v==name)
                    return Term->Copy();
                else
                    return Copy();
            }

            bool reducible()
            { 
                return symbol_table.find(name) != symbol_table.end();
                // delta-reduction possible
            }

            LambdaTerm* reduce()
            {
                return symbol_table[name]->Copy();
                // delta-reduction
            }
    };

    class Function : public LambdaTerm
    {
        public:
            string var;
            LambdaTerm *Term;

            bool contains_free_var(string v)
            {
                if(v == var) // in this case its not a FREE variable
                    return false;
                return Term->contains_free_var(v);
            }

            Function(string v, LambdaTerm *t)
            {
                var=v;
                Term=t;
                t->SetBound(v);
            }
            
            ~Function()
            {
                delete Term;
            }
            
            LambdaTerm* Copy()
            {
                return new Function(var, Term->Copy());
            }

            void SetBound(string vname)
            {
                Term->SetBound(vname);
            }

            string ToString()
            {
                return string(latex_output ? "\\lambda " : "\\")+var+"."+Term->ToString();
            }

            LambdaTerm* substitute(string v, LambdaTerm *term)
            {
                if(v == var)
                    return Copy();
                if(term->contains_free_var(var))
                {
                    Variable* newvar= new Variable(var + "_{" + to_string(replacement_var_id++) + "}");
                    LambdaTerm *temp = Term->substitute(var, newvar);
                    LambdaTerm *result = new Function(newvar->name, temp->substitute(v, term) );
                    delete temp;
                    delete newvar;
                    return result;
                }
                return new Function(var, Term->substitute(v,term));
            }

            bool reducible()
            {
                return Term->reducible();
            }

            LambdaTerm* reduce()
            {
                return new Function(var, Term->reduce());
            }
    };

    class Application : public LambdaTerm
    {
        public:
            LambdaTerm *Funct;
            LambdaTerm *Parameter;

            bool contains_free_var(string v)
            {
                return Funct->contains_free_var(v) || Parameter->contains_free_var(v);
            }

            void SetBound(string vname)
            {
                Funct->SetBound(vname);
                Parameter->SetBound(vname);
            }

            Application(LambdaTerm *f, LambdaTerm *p)
            {
                Funct = f;
                Parameter=p;
            }
            
            ~Application()
            {
                delete Funct;
                delete Parameter;
            }
            
            LambdaTerm* Copy()
            {
                return new Application(Funct->Copy(), Parameter->Copy());
            }

            string ToString()
            {
                return string("(")+Funct->ToString() + (latex_output?"\\: ":" ") + Parameter->ToString()+")";
            }

            LambdaTerm* substitute(string v, LambdaTerm *Term)
            {
                return new Application(Funct->substitute(v,Term), Parameter->substitute(v,Term));
            }

            bool reducible()
            {
                if(dynamic_cast<Function*>(Funct) != NULL) // beta-reduction possible
                    return true;
                return Funct->reducible() || Parameter->reducible();
            }

            LambdaTerm* reduce()
            {
                if(dynamic_cast<Function*>(Funct) != NULL) // beta-reduction
                {
                    Function* f = dynamic_cast<Function*>(Funct);
                    return f->Term->substitute(f->var, Parameter);
                }
                if(Funct->reducible())
                    return new Application(Funct->reduce(), Parameter->Copy());
                return new Application(Funct->Copy(), Parameter->reduce());
            }
    };

    LambdaTerm* IntToLambda(int x)
    {
        if(x <= 0)
            return new Variable("x");
        else
            return new Application(new Variable("f"), IntToLambda(x-1));
    }
    
    void yyerror(const char* s)
    {
        cerr << "SYNTACTIC ERROR (Line " << yylineno << "): " << s << endl;
    }

    void reduce_all(LambdaTerm* t)
    {
        LambdaTerm* temp; int steps = 0;
        if(latex_output)
            os << "& & ";
            
        while(t->reducible())
        {
            if(verbose)
                os << t->ToString() << (latex_output ? "\\\\\n& = & " : "\n\n");
                
            temp = t;
            t = t->reduce();
            delete temp;
            steps++;
        }
        os << t->ToString() << "\n\n(needed steps: " << steps << ")\n\n\n\n";
        delete t;
    }
%}

    //--------------------------------------------------------------------------

%start commands
%union {
    string* text;
    LambdaTerm* term;
}
%token <text> ID EQ LAMBDA ENDLAMBDA OPEN CLOSE INT EOL SEMICOLON
%nonassoc EQ
%nonassoc LAMBDA ENDLAMBDA
%nonassoc INT ID
%left apply
%nonassoc OPEN CLOSE

%%


commands     : commands command
             | 
             ;
             
command      : ID EQ term SEMICOLON
             {
                  if(symbol_table.find(*$<text>1) != symbol_table.end())
                      delete symbol_table[*$<text>1];
                  symbol_table[*$<text>1] = $<term>3;
                  delete $<text>1;
             }
             | ID EQ applications SEMICOLON
             {
                  if(symbol_table.find(*$<text>1) != symbol_table.end())
                      delete symbol_table[*$<text>1];
                  symbol_table[*$<text>1] = $<term>3;
                  delete $<text>1;
             }
             | ID EQ LAMBDA lambda_ids SEMICOLON
             {
                  if(symbol_table.find(*$<text>1) != symbol_table.end())
                      delete symbol_table[*$<text>1];
                  symbol_table[*$<text>1] = $<term>4;
                  delete $<text>1;
             }
             | term SEMICOLON { reduce_all($<term>1); }
             | applications SEMICOLON { reduce_all($<term>1); }
             | LAMBDA lambda_ids SEMICOLON {reduce_all($<term>1);}
             ;
             
applications : applications term {$<term>$ = new Application($<term>1, $<term>2);} %prec apply
             | term term {$<term>$ = new Application($<term>1, $<term>2);} %prec apply
             ;
             
term         : OPEN applications CLOSE {$<term>$ = $<term>2;}
             | ID {$<term>$ = new Variable(*$<text>1); delete $<text>1;}
             | OPEN LAMBDA lambda_ids CLOSE {$<term>$ = $<term>3;}
             | INT {$<term>$ = new Function("f", new Function("x",IntToLambda(atoi($<text>1->c_str())))); delete $<text>1;}
             | OPEN term CLOSE {$<term>$=$<term>2;}
             ;
             
lambda_ids   : ID lambda_ids {$<term>$ = new Function(*$<text>1,$<term>2); delete $<text>1;}
             | ENDLAMBDA applications {$<term>$ = $<term>2;}
             | ENDLAMBDA term {$<term>$ = $<term>2;}
             | ENDLAMBDA LAMBDA lambda_ids {$<term>$ = $<term>3;}
             ;

%%

#ifdef TARGET_PLATFORM_NATIVE
int main(int argc, char** argv)
{
    int i;
    for(i=1; i < argc; i++)
    {
        string argvi = argv[i];
        if(argvi == "-v" || argvi == "--verbose")
            verbose = true;
        else if(argvi == "-e" || argvi == "--expand-names")
            expandnames = true;
        else if(argvi == "-t" || argvi == "--latex")
            latex_output = true;
        else if(argvi == "-l" || argvi == "--no-lazy-evaluation")
            lazy_evaluation = false;
        else if(argvi == "-s" || argvi == "--silent")
            silent = true;
        else if(argvi == "-h" || argvi == "--help")
        {
            os << "\n";
            os << " pure \xce\xbb-calculus interpreter\n";
            os << " \xc2\xa9 2009 by Viktor \"AlgorithMan\" Engelmann\n";
            os << "                 www.AlgorithMan.de\n";
            os << " Distributed under the terms of the GNU general\n";
            os << " public license (GPL) 3. See gpl3.txt for details\n";
            os << "\n";
            os << " usage: " << argv[0] << " [options] [sourcecodes]\n";
            os << " where options can be\n";
            os << "      -e  --expand-names\n";
            os << "         to print the functions\'s codes instead of their names\n";
            os << "         (generates pure \xce\xbb-terms as output)\n";
            os << "      -v --verbose\n";
            os << "         prints intermediate steps\n";
            os << "      -t --latex\n";
            os << "         generates LaTeX compatible output\n";
            os << "      -s --silent\n";
            os << "         suppress messages\n";
            os << "      -h --help\n";
            os << "         this help message\n";
            os << " example sourcecode:\n";
            os << "      succ = \\n.\\f.\\x.f(n f x);\n";
            os << "      add = \\n m.n succ m;\n";
            os << " example program call\n";
            os << "      "<<argv[0]<<" -v stdfunctions.plc\n";
            os << " example query\n";
            os << "      add 1 1;\n";
            os << "      (result: \\f.\\x.(f (f x)) which is the \xce\xbb-calculus representation of 2)\n";
            os << " if you don\'t understand this whole \xce\xbb-calculus stuff, read\n";
            os << "      http://en.wikipedia.org/wiki/Lambda_calculus\n";
            os << "      http://en.wikipedia.org/wiki/Fixed_point_combinator\n";
            os << "\n";
            return 0;
        }
        else
            break;
    }

    if(!silent)
    {
        cerr << "\n";
        cerr << "pure \xce\xbb-calculus interpreter\n";
        cerr << "\xc2\xa9 2009 by Viktor \"AlgorithMan\" Engelmann\n";
        cerr << "                www.AlgorithMan.de\n";
        cerr << "Distributed under the terms of the GNU general\n";
        cerr << "public license (GPL) 3. See gpl3.txt for details\n";
        cerr << "\n";
    }


    for(; i < argc; i++)
    {
        if(!silent)
            cerr << "loading file \"" << argv[i] << "\"...";
        yyin = fopen(argv[i], "r");
        yyparse();
        fclose(yyin);
        if(!silent)
            cerr << " done\n";
    }
    if(!silent)
        cerr << "\n";
    cerr.flush();

    if(latex_output)
    {
        os << "\\documentclass[a4paper,10pt,landscape]{article}\n";
        os << "\\usepackage{array}\n";
        os << "\\usepackage{hyperref}\n";
        os << "\n";
        os << "\\begin{document}\n";
        os << "Output generated with pure $\\lambda$-calculus interpreter \\textbf{plc}\\\\\n";
        os << "\\copyright 2009 by Viktor \"AlgorithMan\" Engelmann \\url{http://www.AlgorithMan.de}\n";
    }

    yyin = stdin;
    yyparse();

    if(latex_output)
        os << "\\end{document}\n\n";

    for(map<string,LambdaTerm*>::iterator it = symbol_table.begin(); it != symbol_table.end(); it++)
        delete it->second;

    return 0;
}
#endif // TARGET_PLATFORM_NATIVE

#ifdef TARGET_PLATFORM_WASM
extern "C" double RunPureLambdaCalculus(char* code, double flags)
{
    int iflags = (int)flags;
    //DEBUG_LEXER = ((iflags&1) != 0);
    //DEBUG_PARSER = ((iflags&2) != 0);
    verbose = true;

    try
    {
        YY_BUFFER_STATE bufferstate = yy_scan_string(code);
        yyparse();
        yy_delete_buffer(bufferstate);
    }
    catch(const char* ex)
    {
        cerr << " Fatal Exception: " << ex << "\n\n";
    }

    EM_ASM_ARGS({
        var msg = Pointer_stringify($0);
        ShowResult(msg);
    }, os.str().c_str());

    for(map<string,LambdaTerm*>::iterator it = symbol_table.begin(); it != symbol_table.end(); it++)
        delete it->second;

    return 0;
}
#endif // TARGET_PLATFORM_WASM


